{ Copyright (C) 1998-2023, written by Shkolnik Mike, Scalabium
  E-Mail:  mshkolnik@scalabium.com
           mshkolnik@yahoo.com
  WEB: http://www.scalabium.com

  This component allow to display and modify the graphical star rating.
  The rating could be displayed with stars and part of star.
  Instead default stars you may load the own custom graphic (bullets, for example)
}
unit SMStarRate;

interface

uses
  Windows, Messages, Classes, Graphics, Controls;

type
  TSMStarDrawMode = (sdmFull, sdmHalf, sdmEmpty);

  {$IFDEF SM_ADD_ComponentPlatformsAttribute}
  [ComponentPlatformsAttribute(pidWin32 or pidWin64)]
  {$ENDIF}
  TSMStarRating = class(TGraphicControl)
  private
    { Private declarations }
    FRating: Double;
    FMin: Integer;
    FMax: Integer;

    FMargin: Integer;
    FSpacing: Integer;

    FStarColor: TColor;

    FReadOnly: Boolean;

    FAutoSize: Boolean;

    procedure AdjustBounds;
    function GetTransparent: Boolean;
    procedure SetTransparent(Value: Boolean);

    procedure SetRating(Value: Double);
    procedure SetMin(Value: Integer);
    procedure SetMax(Value: Integer);
    procedure SetMargin(Value: Integer);
    procedure SetSpacing(Value: Integer);
    procedure SetStarColor(Value: TColor);
  protected
    { Protected declarations }
    procedure Loaded; override;
    procedure Paint; override;
    procedure SetHLAutoSize(Value: Boolean); virtual;

    procedure DrawStar(ACanvas: TCanvas; ARect: TRect; StartIndex: Integer; DrawMode: TSMStarDrawMode);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;

    property Canvas;
  published
    { Published declarations }
    property Rating: Double read FRating write SetRating;
    property Min: Integer read FMin write SetMin default 0;
    property Max: Integer read FMax write SetMax default 5;

    property ReadOnly: Boolean read FReadOnly write FReadOnly default True;

    property Margin: Integer read FMargin write SetMargin default 2;
    property Spacing: Integer read FSpacing write SetSpacing default 5;

    property StarColor: TColor read FStarColor write SetStarColor default clYellow;

    property AutoSize: Boolean read FAutoSize write SetHLAutoSize default False;
    property Transparent: Boolean read GetTransparent write SetTransparent default False;

    property Align;
    property Color;
    property DragCursor;
    property DragMode;
    property Enabled;
    property ParentColor;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

implementation
{$R *.RES}

uses SysUtils;

{ TSMStarRating }
constructor TSMStarRating.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FRating := 0;
  FMin := 0;
  FMax := 5;
  FReadOnly := True;
  FMargin := 2;
  FSpacing := 5;
  FStarColor := clYellow;

  ControlStyle := ControlStyle + [csOpaque, csReplicatable];
  Width := 120;
  Height := 20;
  FAutoSize := False;
end;

procedure TSMStarRating.DrawStar(ACanvas: TCanvas; ARect: TRect; StartIndex: Integer; DrawMode: TSMStarDrawMode);
var
  arrStar: array[0..9] of TPoint;
  rectWidth, rectHeight: Integer;
begin
  if not Assigned(ACanvas) then exit;
  
  rectWidth := ARect.Right - ARect.Left;
  rectHeight := ARect.Bottom - ARect.Top;

  if (DrawMode = sdmFull) and (FStarColor <> Color) then
    ACanvas.Brush.Color := FStarColor
  else
    ACanvas.Brush.Color := Color;

  arrStar[0].X := ARect.Left + (rectWidth div 2);
  arrStar[0].Y := ARect.Top;
  arrStar[1].X := ARect.Left + (42*rectWidth div 64);
  arrStar[1].Y := ARect.Top + (19*rectHeight div 64);
  arrStar[2].X := ARect.Left + rectWidth;
  arrStar[2].Y := ARect.Top + (22*rectHeight div 64);
  arrStar[3].X := ARect.Left + (48*rectWidth div 64);
  arrStar[3].Y := ARect.Top + (38*rectHeight div 64);
  arrStar[4].X := ARect.Left + (52*rectWidth div 64);
  arrStar[4].Y := ARect.Top + rectHeight;
  arrStar[5].X := ARect.Left + (rectWidth div 2);
  arrStar[5].Y := ARect.Top + (52*rectHeight div 64);
  arrStar[6].X := ARect.Left + (12*rectWidth div 64);
  arrStar[6].Y := ARect.Top + rectHeight;
  arrStar[7].X := ARect.Left + (rectWidth div 4);
  arrStar[7].Y := ARect.Top + (38*rectHeight div 64);
  arrStar[8].X := ARect.Left;
  arrStar[8].Y := ARect.Top + (22*rectHeight div 64);
  arrStar[9].X := ARect.Left + (22*rectWidth div 64);
  arrStar[9].Y := ARect.Top + (19*rectHeight div 64);

  ACanvas.Polygon(arrStar);

  if (DrawMode = sdmHalf) and (FStarColor <> Color) then
  begin
    ACanvas.Brush.Color := FStarColor;
    ACanvas.Polygon([arrStar[0], arrStar[5], arrStar[6], arrStar[7], arrStar[8], arrStar[9]]);
  end
end;

procedure TSMStarRating.Paint;
var
  ARect: TRect;

  i, StarWidth, StarHeight: Integer;
//  StarAreas: array of TRect;
begin
//  if (csDestroying in ComponentState) then exit;

  with Canvas do
  begin
    if not Transparent then
    begin
      Brush.Color := Self.Color;
      Brush.Style := bsSolid;
      FillRect(ClientRect);
    end;
    Brush.Style := bsClear;

    StarWidth := (ClientWidth - (2*FMargin + FSpacing*(FMax - FMin))) div (FMax - FMin);
    StarHeight := ClientHeight - 2*FMargin;


    ARect.Left := Margin;
    ARect.Top := Margin;
    ARect.Right := StarWidth;
    ARect.Bottom := StarHeight;

//    SetLength(StarAreas, FMax-FMin-1);
    for i := FMin to FMax-1 do
    begin
(*
      StarAreas[i-FMin].Left := ARect.Left - (FSpacing div 2);
      StarAreas[i-FMin].Top := ARect.Top;

      StarAreas[i-FMin].Right := StarAreas[i-FMin].Left + ARect.Right - ARect.Left + (FSpacing div 2);
      StarAreas[i-FMin].Bottom := StarAreas[i-FMin].Top + StarHeight;
*)
      if (i+1 <= FRating) then
        DrawStar(Canvas, ARect, i, sdmFull)
      else
      if (i < Trunc(FRating + 0.5)) then
        DrawStar(Canvas, ARect, i, sdmHalf)
      else
        DrawStar(Canvas, ARect, i, sdmEmpty);

      ARect.Left := ARect.Left + StarWidth + FSpacing;
      ARect.Right := ARect.Left + StarWidth;
    end;
  end;
end;

procedure TSMStarRating.Loaded;
begin
  inherited Loaded;
  AdjustBounds;
end;

procedure TSMStarRating.AdjustBounds;
begin
  if not (csReading in ComponentState) and FAutoSize then
  begin
    SetBounds(Left, Top, 2*FMargin + FSpacing*(FMax - FMin - 1), Height);
  end;
end;

procedure TSMStarRating.SetHLAutoSize(Value: Boolean);
begin
  if FAutoSize <> Value then
  begin
    FAutoSize := Value;
    AdjustBounds;
  end;
end;

function TSMStarRating.GetTransparent: Boolean;
begin
  Result := not (csOpaque in ControlStyle);
end;

procedure TSMStarRating.SetTransparent(Value: Boolean);
begin
  if Transparent <> Value then
  begin
    if Value then
      ControlStyle := ControlStyle - [csOpaque]
    else
      ControlStyle := ControlStyle + [csOpaque];
    Invalidate;
  end;
end;

procedure TSMStarRating.SetRating(Value: Double);
begin
  if (FRating <> Value) then
  begin
    FRating := Value;
    Invalidate;
  end;
end;

procedure TSMStarRating.SetMin(Value: Integer);
begin
  if (FMin <> Value) then
  begin
    FMin := Value;
    Invalidate;
  end;
end;

procedure TSMStarRating.SetMax(Value: Integer);
begin
  if (FMax <> Value) then
  begin
    FMax := Value;
    Invalidate;
  end;
end;

procedure TSMStarRating.SetMargin(Value: Integer);
begin
  if (FMargin <> Value) then
  begin
    FMargin := Value;
    Invalidate;
  end;
end;

procedure TSMStarRating.SetSpacing(Value: Integer);
begin
  if (FSpacing <> Value) then
  begin
    FSpacing := Value;
    Invalidate;
  end;
end;

procedure TSMStarRating.SetStarColor(Value: TColor);
begin
  if (FStarColor <> Value) then
  begin
    FStarColor := Value;
    Invalidate;
  end;
end;

procedure TSMStarRating.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  StarWidth: Integer;
begin
  inherited MouseDown(Button, Shift, X, Y);

  if not ReadOnly then
  begin
    {calculate in what part (star) of control we clicked}
    StarWidth := (ClientWidth - (2*FMargin + FSpacing*(FMax - FMin))) div (FMax - FMin);

    Rating := (X+StarWidth div 2) div StarWidth
  end;
end;

end.
